@@ -14,7 +14,7 @@
 #
 # You should have received a copy of the GNU Lesser General Public License
 # along with 'ssh'; if not, write to the Free Software Foundation, Inc.,
-# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
+# 51 Franklin Street, Suite 500, Boston, MA  02110-1335  USA.
 
 """
 SSH Agent interface for Unix clients.
@@ -28,7 +28,6 @@ import threading
 import time
 import tempfile
 import stat
-import fcntl
 from select import select
 
 from ssh.ssh_exception import SSHException
@@ -36,6 +35,7 @@ from ssh.message import Message
 from ssh.pkey import PKey
 from ssh.channel import Channel
 from ssh.common import io_sleep
+from ssh.util import retry_on_signal
 
 SSH2_AGENTC_REQUEST_IDENTITIES, SSH2_AGENT_IDENTITIES_ANSWER, \
     SSH2_AGENTC_SIGN_REQUEST, SSH2_AGENT_SIGN_RESPONSE = range(11, 15)
@@ -46,7 +46,7 @@ class AgentSSH(object):
     local machine.  If an SSH agent is running, this class can be used to
     connect to it and retreive L{PKey} objects which can be used when
     attempting to authenticate to remote SSH servers.
-    
+
     Because the SSH agent protocol uses environment variables and unix-domain
     sockets, this probably doesn't work on Windows.  It does work on most
     posix platforms though (Linux and MacOS X, for example).
@@ -60,7 +60,7 @@ class AgentSSH(object):
         Return the list of keys available through the SSH agent, if any.  If
         no SSH agent was running (or it couldn't be contacted), an empty list
         will be returned.
-        
+
         @return: a list of keys available on the SSH agent
         @rtype: tuple of L{AgentKey}
         """
@@ -119,6 +119,7 @@ class AgentProxyThread(threading.Thread):
             raise
 
     def _communicate(self):
+        import fcntl
         oldflags = fcntl.fcntl(self.__inr, fcntl.F_GETFL)
         fcntl.fcntl(self.__inr, fcntl.F_SETFL, oldflags | os.O_NONBLOCK)
         while not self._exit:
@@ -186,22 +187,23 @@ class AgentClientProxy(object):
        the remote fake agent and the local agent
        -> Communication occurs ...
     """
-    def __init__(self, chanClient):
+    def __init__(self, chanRemote):
         self._conn = None
-        self.__chanC = chanClient
-        chanClient.request_forward_agent(self._forward_agent_handler)
+        self.__chanR = chanRemote
+        self.thread = AgentRemoteProxy(self, chanRemote)
+        self.thread.start()
 
     def __del__(self):
         self.close()
 
     def connect(self):
-        """ 
+        """
         Method automatically called by the run() method of the AgentProxyThread
         """
         if ('SSH_AUTH_SOCK' in os.environ) and (sys.platform != 'win32'):
             conn = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
             try:
-                conn.connect(os.environ['SSH_AUTH_SOCK'])
+                retry_on_signal(lambda: conn.connect(os.environ['SSH_AUTH_SOCK']))
             except:
                 # probably a dangling env var: the ssh agent is gone
                 return
@@ -227,10 +229,6 @@ class AgentClientProxy(object):
         if self._conn is not None:
             self._conn.close()
 
-    def _forward_agent_handler(self, chanRemote):
-        self.thread = AgentRemoteProxy(self, chanRemote)
-        self.thread.start()
-
 class AgentServerProxy(AgentSSH):
     """
     @param t : transport used for the Forward for SSH Agent communication
@@ -274,31 +272,48 @@ class AgentServerProxy(AgentSSH):
         @return: the SSH_AUTH_SOCK Environnement variables
         @rtype: dict
         """
-        env = {} 
+        env = {}
         env['SSH_AUTH_SOCK'] = self._get_filename()
         return env
 
     def _get_filename(self):
         return self._file
 
+class AgentRequestHandler(object):
+    def __init__(self, chanClient):
+        self._conn = None
+        self.__chanC = chanClient
+        chanClient.request_forward_agent(self._forward_agent_handler)
+        self.__clientProxys = []
+
+    def _forward_agent_handler(self, chanRemote):
+        self.__clientProxys.append(AgentClientProxy(chanRemote))
+
+    def __del__(self):
+        self.close()
+
+    def close(self):
+        for p in self.__clientProxys:
+            p.close()
+
 class Agent(AgentSSH):
     """
     Client interface for using private keys from an SSH agent running on the
     local machine.  If an SSH agent is running, this class can be used to
     connect to it and retreive L{PKey} objects which can be used when
     attempting to authenticate to remote SSH servers.
-    
+
     Because the SSH agent protocol uses environment variables and unix-domain
     sockets, this probably doesn't work on Windows.  It does work on most
     posix platforms though (Linux and MacOS X, for example).
     """
-    
+
     def __init__(self):
         """
         Open a session with the local machine's SSH agent, if one is running.
         If no agent is running, initialization will succeed, but L{get_keys}
         will return an empty tuple.
-        
+
         @raise SSHException: if an SSH agent is found, but speaks an
             incompatible protocol
         """
@@ -320,7 +335,7 @@ class Agent(AgentSSH):
         else:
             # no agent support
             return
-        self._connect(conn)            
+        self._connect(conn)
 
     def close(self):
         """
@@ -334,7 +349,7 @@ class AgentKey(PKey):
     authenticating to a remote server (signing).  Most other key operations
     work as expected.
     """
-    
+
     def __init__(self, agent, blob):
         self.agent = agent
         self.blob = blob
