@@ -14,7 +14,7 @@
 #
 # You should have received a copy of the GNU Lesser General Public License
 # along with 'ssh'; if not, write to the Free Software Foundation, Inc.,
-# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA.
+# 51 Franklin Street, Suite 500, Boston, MA  02110-1335  USA.
 
 """
 L{Transport} handles the core SSH2 protocol.
@@ -45,6 +45,7 @@ from ssh.rsakey import RSAKey
 from ssh.server import ServerInterface
 from ssh.sftp_client import SFTPClient
 from ssh.ssh_exception import SSHException, BadAuthenticationType, ChannelException
+from ssh.util import retry_on_signal
 
 from Crypto import Random
 from Crypto.Cipher import Blowfish, AES, DES3, ARC4
@@ -289,7 +290,7 @@ class Transport (threading.Thread):
                     addr = sockaddr
                     sock = socket.socket(af, socket.SOCK_STREAM)
                     try:
-                        sock.connect((hostname, port))
+                        retry_on_signal(lambda: sock.connect((hostname, port)))
                     except socket.error, e:
                         reason = str(e)
                     else:
@@ -1529,6 +1530,10 @@ class Transport (threading.Thread):
         # indefinitely, creating a GC cycle and not letting Transport ever be
         # GC'd. it's a bug in Thread.)
 
+        # Hold reference to 'sys' so we can test sys.modules to detect
+        # interpreter shutdown.
+        self.sys = sys
+
         # Required to prevent RNG errors when running inside many subprocess
         # containers.
         Random.atfork()
@@ -1540,94 +1545,102 @@ class Transport (threading.Thread):
         else:
             self._log(DEBUG, 'starting thread (client mode): %s' % hex(long(id(self)) & 0xffffffffL))
         try:
-            self.packetizer.write_all(self.local_version + '\r\n')
-            self._check_banner()
-            self._send_kex_init()
-            self._expect_packet(MSG_KEXINIT)
+            try:
+                self.packetizer.write_all(self.local_version + '\r\n')
+                self._check_banner()
+                self._send_kex_init()
+                self._expect_packet(MSG_KEXINIT)
 
-            while self.active:
-                if self.packetizer.need_rekey() and not self.in_kex:
-                    self._send_kex_init()
-                try:
-                    ptype, m = self.packetizer.read_message()
-                except NeedRekeyException:
-                    continue
-                if ptype == MSG_IGNORE:
-                    continue
-                elif ptype == MSG_DISCONNECT:
-                    self._parse_disconnect(m)
-                    self.active = False
-                    self.packetizer.close()
-                    break
-                elif ptype == MSG_DEBUG:
-                    self._parse_debug(m)
-                    continue
-                if len(self._expected_packet) > 0:
-                    if ptype not in self._expected_packet:
-                        raise SSHException('Expecting packet from %r, got %d' % (self._expected_packet, ptype))
-                    self._expected_packet = tuple()
-                    if (ptype >= 30) and (ptype <= 39):
-                        self.kex_engine.parse_next(ptype, m)
+                while self.active:
+                    if self.packetizer.need_rekey() and not self.in_kex:
+                        self._send_kex_init()
+                    try:
+                        ptype, m = self.packetizer.read_message()
+                    except NeedRekeyException:
                         continue
-
-                if ptype in self._handler_table:
-                    self._handler_table[ptype](self, m)
-                elif ptype in self._channel_handler_table:
-                    chanid = m.get_int()
-                    chan = self._channels.get(chanid)
-                    if chan is not None:
-                        self._channel_handler_table[ptype](chan, m)
-                    elif chanid in self.channels_seen:
-                        self._log(DEBUG, 'Ignoring message for dead channel %d' % chanid)
-                    else:
-                        self._log(ERROR, 'Channel request for unknown channel %d' % chanid)
+                    if ptype == MSG_IGNORE:
+                        continue
+                    elif ptype == MSG_DISCONNECT:
+                        self._parse_disconnect(m)
                         self.active = False
                         self.packetizer.close()
-                elif (self.auth_handler is not None) and (ptype in self.auth_handler._handler_table):
-                    self.auth_handler._handler_table[ptype](self.auth_handler, m)
+                        break
+                    elif ptype == MSG_DEBUG:
+                        self._parse_debug(m)
+                        continue
+                    if len(self._expected_packet) > 0:
+                        if ptype not in self._expected_packet:
+                            raise SSHException('Expecting packet from %r, got %d' % (self._expected_packet, ptype))
+                        self._expected_packet = tuple()
+                        if (ptype >= 30) and (ptype <= 39):
+                            self.kex_engine.parse_next(ptype, m)
+                            continue
+
+                    if ptype in self._handler_table:
+                        self._handler_table[ptype](self, m)
+                    elif ptype in self._channel_handler_table:
+                        chanid = m.get_int()
+                        chan = self._channels.get(chanid)
+                        if chan is not None:
+                            self._channel_handler_table[ptype](chan, m)
+                        elif chanid in self.channels_seen:
+                            self._log(DEBUG, 'Ignoring message for dead channel %d' % chanid)
+                        else:
+                            self._log(ERROR, 'Channel request for unknown channel %d' % chanid)
+                            self.active = False
+                            self.packetizer.close()
+                    elif (self.auth_handler is not None) and (ptype in self.auth_handler._handler_table):
+                        self.auth_handler._handler_table[ptype](self.auth_handler, m)
+                    else:
+                        self._log(WARNING, 'Oops, unhandled type %d' % ptype)
+                        msg = Message()
+                        msg.add_byte(chr(MSG_UNIMPLEMENTED))
+                        msg.add_int(m.seqno)
+                        self._send_message(msg)
+            except SSHException, e:
+                self._log(ERROR, 'Exception: ' + str(e))
+                self._log(ERROR, util.tb_strings())
+                self.saved_exception = e
+            except EOFError, e:
+                self._log(DEBUG, 'EOF in transport thread')
+                #self._log(DEBUG, util.tb_strings())
+                self.saved_exception = e
+            except socket.error, e:
+                if type(e.args) is tuple:
+                    emsg = '%s (%d)' % (e.args[1], e.args[0])
                 else:
-                    self._log(WARNING, 'Oops, unhandled type %d' % ptype)
-                    msg = Message()
-                    msg.add_byte(chr(MSG_UNIMPLEMENTED))
-                    msg.add_int(m.seqno)
-                    self._send_message(msg)
-        except SSHException, e:
-            self._log(ERROR, 'Exception: ' + str(e))
-            self._log(ERROR, util.tb_strings())
-            self.saved_exception = e
-        except EOFError, e:
-            self._log(DEBUG, 'EOF in transport thread')
-            #self._log(DEBUG, util.tb_strings())
-            self.saved_exception = e
-        except socket.error, e:
-            if type(e.args) is tuple:
-                emsg = '%s (%d)' % (e.args[1], e.args[0])
-            else:
-                emsg = e.args
-            self._log(ERROR, 'Socket exception: ' + emsg)
-            self.saved_exception = e
-        except Exception, e:
-            self._log(ERROR, 'Unknown exception: ' + str(e))
-            self._log(ERROR, util.tb_strings())
-            self.saved_exception = e
-        _active_threads.remove(self)
-        for chan in self._channels.values():
-            chan._unlink()
-        if self.active:
-            self.active = False
-            self.packetizer.close()
-            if self.completion_event != None:
-                self.completion_event.set()
-            if self.auth_handler is not None:
-                self.auth_handler.abort()
-            for event in self.channel_events.values():
-                event.set()
-            try:
-                self.lock.acquire()
-                self.server_accept_cv.notify()
-            finally:
-                self.lock.release()
-        self.sock.close()
+                    emsg = e.args
+                self._log(ERROR, 'Socket exception: ' + emsg)
+                self.saved_exception = e
+            except Exception, e:
+                self._log(ERROR, 'Unknown exception: ' + str(e))
+                self._log(ERROR, util.tb_strings())
+                self.saved_exception = e
+            _active_threads.remove(self)
+            for chan in self._channels.values():
+                chan._unlink()
+            if self.active:
+                self.active = False
+                self.packetizer.close()
+                if self.completion_event != None:
+                    self.completion_event.set()
+                if self.auth_handler is not None:
+                    self.auth_handler.abort()
+                for event in self.channel_events.values():
+                    event.set()
+                try:
+                    self.lock.acquire()
+                    self.server_accept_cv.notify()
+                finally:
+                    self.lock.release()
+            self.sock.close()
+        except:
+            # Don't raise spurious 'NoneType has no attribute X' errors when we
+            # wake up during interpreter shutdown. Or rather -- raise
+            # everything *if* sys.modules (used as a convenient sentinel)
+            # appears to still exist.
+            if self.sys.modules is not None:
+                raise
 
 
     ###  protocol stages
@@ -2008,8 +2021,8 @@ class Transport (threading.Thread):
         initial_window_size = m.get_int()
         max_packet_size = m.get_int()
         reject = False
-	if (kind == 'auth-agent@openssh.com') and (self._forward_agent_handler is not None):
-	    self._log(DEBUG, 'Incoming forward agent connection')
+        if (kind == 'auth-agent@openssh.com') and (self._forward_agent_handler is not None):
+            self._log(DEBUG, 'Incoming forward agent connection')
             self.lock.acquire()
             try:
                 my_chanid = self._next_channel()
@@ -2087,7 +2100,7 @@ class Transport (threading.Thread):
         m.add_int(self.max_packet_size)
         self._send_message(m)
         self._log(INFO, 'Secsh channel %d (%s) opened.', my_chanid, kind)
-	if kind == 'auth-agent@openssh.com':
+        if kind == 'auth-agent@openssh.com':
             self._forward_agent_handler(chan)
         elif kind == 'x11':
             self._x11_handler(chan, (origin_addr, origin_port))
